import random
import pandas as pd
from sklearn.cluster import KMeans
from sklearn.feature_selection import VarianceThreshold
from sklearn.metrics.pairwise import pairwise_distances
from sklearn.preprocessing import MinMaxScaler
from pyclustering.cluster.kmedoids import kmedoids
import networkx as nx
import multiprocessing as mp
from tsfresh import extract_features
from FeatTS import SLPA
import time
import numpy as np
from itertools import combinations
import os

matrixSym = []


def adaptTimeSeriesUCR(input_data):
    # Initialize empty lists to store the data
    id_list = []
    time_list = []
    value_list = []

    # Loop through the input data
    for i, sublist in enumerate(input_data):
        for j, value in enumerate(sublist):
            id_list.append(i)
            time_list.append(j + 1)  # Adding 1 to start time from 1
            value_list.append(value)

    # Create a DataFrame
    data = {'id': id_list, 'time': time_list, 'value': value_list}
    df = pd.DataFrame(data)
    return df


def choose_and_exclude_indices_by_percentage(classes, percentage):
    # Create a dictionary to store the chosen indices for each class
    class_indices = {}

    # Loop through the classes and assign indices
    for i, class_val in enumerate(classes):
        if class_val not in class_indices:
            class_indices[class_val] = []
        class_indices[class_val].append(i)

    # Determine the number of indices to choose for each class based on the percentage
    chosen_indices = []
    excluded_indices = []
    for class_val, indices in class_indices.items():
        num_indices_to_choose = int(len(indices) * percentage)
        random.shuffle(indices)
        chosen_indices.extend(indices[:num_indices_to_choose])
        excluded_indices.extend(indices[num_indices_to_choose:])

    # Sort the chosen and excluded indices
    chosen_indices.sort()
    excluded_indices.sort()

    return chosen_indices, excluded_indices


def getDataframeAcc(appSeries,perc):
    listClassExtr = list(appSeries.drop_duplicates())
    series = appSeries
    allAccInd = []
    allNotAccInd = []
    for x in listClassExtr:
        sommaClasse = sum(list(series.str.count(x)))
        accepted = int(sommaClasse * perc)
        listIndexAccepted = []
        listIndexNotAccepted = []
        for i in range(len(series)):
            if series[i] == x:
                if len(listIndexAccepted) <= accepted:
                    listIndexAccepted.append(i)
                    allAccInd.append(i)
                else:
                    listIndexNotAccepted.append(i)
                    allNotAccInd.append(i)
    return list(sorted(allAccInd)),list(sorted(allNotAccInd))

def extractFeature(listOut, external_feat=None, n_jobs=1):

    features_filtered_direct = extract_features(listOut, column_id='id', column_sort='time', n_jobs=n_jobs)
    if external_feat is not None:
        features_filtered_direct = features_filtered_direct.join(external_feat)
    features_filtered_direct = normalization_data(features_filtered_direct)

    return features_filtered_direct


def normalization_data(features_filtered_direct):
    features_filtered_direct = features_filtered_direct.dropna(axis='columns')
    # Initialize the MinMaxScaler
    scaler = MinMaxScaler()
    # Fit and transform the DataFrame using the scaler
    normalized_data = scaler.fit_transform(features_filtered_direct)
    # Create a new DataFrame with the normalized data
    features_filtered_direct = pd.DataFrame(normalized_data, columns=features_filtered_direct.columns)
    # Calculate variance for each column
    variance = features_filtered_direct.var()

    # Sort the Series based on values
    sorted_data = variance.sort_values()

    # Reshape the sorted Series into a 2D array
    X = sorted_data.values.reshape(-1, 1)

    # Perform K-Means clustering with k=2
    kmeans = KMeans(n_clusters=2)
    kmeans.fit(X)
    variance_useless_column = variance[:(list(kmeans.labels_).count(0)-1)].index
    features_filtered_direct.drop(columns=variance_useless_column, inplace=True)


    return features_filtered_direct



from concurrent.futures import ProcessPoolExecutor

def getTabNonSym(setCluster, listId, n_jobs=1):
    w = len(listId)
    matrixSym = np.zeros((w, w))

    def matrixCalcParal(result):
        for val in result:
            matrixSym[val["i"], val["j"]] = val["value"]

    totRig = len(listId) // n_jobs

    with ProcessPoolExecutor(max_workers=n_jobs) as executor:
        futures = []
        for i in range(n_jobs):
            start = i * totRig
            end = start + totRig if i < n_jobs - 1 else len(listId)
            futures.append(executor.submit(getValueMatrix, start, listId, end - start, setCluster))
        
        for future in futures:
            matrixCalcParal(future.result())

    for i in range(w):
        maxVal = np.max(matrixSym[i])
        matrixSym[i] = np.abs(matrixSym[i] - maxVal)

    return matrixSym


def getValueMatrix(start,listId,totRig,listOfClust):
    try:
        dictOfValueIJ = []
        # Pre-calcola i risultati di numOfClusterPres per evitare chiamate ripetute
        cluster_pres_cache = {listId[i+start]: numOfClusterPres(listOfClust, listId[i+start]) for i in range(totRig)}

        for i in range(totRig):
            id_i = listId[i + start]
            resultCluster = cluster_pres_cache[id_i]
            for j in range(len(listId)):
                id_j = listId[j]
                resultCouple = numOfRipetitionCouple(id_i, id_j, listOfClust)
                if resultCluster[1] == resultCouple[1]:
                    value = 1
                elif resultCouple[1] == 0:
                    value = 0
                else:
                    value = resultCouple[0] / resultCluster[0]

                dictSingle = {"value": value, "i": i + start, "j": j}
                dictOfValueIJ.append(dictSingle)
        return dictOfValueIJ
    except Exception as e:
        print("Exception in getValueMatrix:")
        print(e)


def getCluster(matrixsym,setCluster,numClust):

    dictTotal = {}
    for x in setCluster:
        listOfDist = []
        for y in setCluster:
            if x != y:
                dictSing = {"id":y,"distance":matrixsym[x][y]}
                listOfDist.append(dictSing)
        dictTotal[x] = listOfDist


    idChoose = getInitialIndex(dictTotal,numClust)
    D = pairwise_distances(matrixsym, metric='correlation')


    kmedoids_instance = kmedoids(D, idChoose, tolerance=0.000001)
    kmedoids_instance.process()
    Cl = kmedoids_instance.get_clusters()
    # show allocated clusters

    dictClu = {}
    for i in range(0, len(Cl)):
        dictApp = {i: Cl[i]}
        dictClu.update(dictApp)


    listOfCommFind = []

    for label in dictClu:
        for point_idx in dictClu[label]:
            dictSing = {"label": label, "cluster": list(setCluster)[point_idx]}
            listOfCommFind.append(dictSing)
    return listOfCommFind

def numOfClusterPres(setCluster,id):
    countId = 0
    countTimes = 0
    for i in range(0,len(setCluster)):
        if id in (setCluster[i]["list"]):
            countId += (setCluster[i]["weight"])
            countTimes += 1
    return countId, countTimes

def numOfRipetitionCouple(id1,id2,setCluster):
    countId = 0
    countTimes = 0
    for i in range(0,len(setCluster)):
        if id1 in (setCluster[i]["list"]) and id2 in (setCluster[i]["list"]):
            countId += setCluster[i]["weight"]
            countTimes += 1
    return countId,countTimes


def listOfId(setCluster):
    listId = set()
    for value in setCluster:
        for id in value:
            listId.add(id)
    return list(listId)

def createSet(listOfCommFind,clusterK):
    listOfCluster = []
    for i in range(0,clusterK):
        dictSing = {"cluster":[],"label":i}
        listOfCluster.append(dictSing)
    for value in listOfCommFind:
        listApp = listOfCluster[value["label"]]["cluster"]
        listApp.append(value["cluster"])
        listOfCluster.remove(listOfCluster[value["label"]])
        dictSing = {"cluster":listApp,"label":value["label"]}
        listOfCluster.insert(value["label"],dictSing)

    return listOfCluster


def randomFeat(ris,numberFeatUse):

    ris = ris.dropna(subset=['p_value'])

    indexNames = ris[ris['relevant'] == True].index

    ris.drop(indexNames, inplace=True)
    randomFeat = ris.sample(n=numberFeatUse)


    return randomFeat


def getCommunityDetectionTrain(features, features_filtered_direct, listOfId, threshold, clusterK, chooseAlgorithm):
    dictOfInfo = {}

    for feature in features:
        G = nx.Graph()
        H = nx.path_graph(listOfId)
        G.add_nodes_from(H)

        # Convert the feature column to a numpy array for efficient operations
        values = features_filtered_direct[feature].to_numpy()
        # Calculate the pairwise absolute differences using broadcasting
        diff_matrix = np.abs(values[:, np.newaxis] - values)
        # Extract the upper triangle of the matrix, excluding the diagonal
        listOfDistance = diff_matrix[np.triu_indices(len(values), k=1)]
        # Find the threshold index
        threshold_index = int(len(listOfDistance) * threshold)
        # Use numpy's partition to find the threshold value without full sorting
        distanceMinAccept = np.partition(listOfDistance, threshold_index)[threshold_index]

        for i, j in combinations(range(len(listOfId)), 2):
            if diff_matrix[i, j] < distanceMinAccept:
                G.add_edge(i, j)
        
        try:
            if list(chooseAlgorithm.keys())[0] == 'SLPA':
                extrC = SLPA.find_communities(G, chooseAlgorithm['SLPA']['iteration'], chooseAlgorithm['SLPA']['radious'])
                coms = []
                for val in extrC:
                    coms.append(frozenset(extrC[val]))
            elif list(chooseAlgorithm.keys())[0] == 'kClique':
                coms = list(nx.algorithms.community.k_clique_communities(G, chooseAlgorithm['SLPA']['trainClique']))
            elif list(chooseAlgorithm.keys())[0] == 'Greedy':
                coms = list(nx.algorithms.community.greedy_modularity_communities(G))

            if len(coms) > clusterK:
                dictOfInfo[feature] = {"distance": distanceMinAccept, "cluster": coms, "weightFeat": clusterK / len(coms)}
            else:
                dictOfInfo[feature] = {"distance": distanceMinAccept, "cluster": coms, "weightFeat": len(coms) / clusterK}
            
        except Exception as e:
            print(e)
            pass
    return dictOfInfo


def cleaning(df: pd.DataFrame) -> pd.DataFrame:
    # Drop columns which contain nan, +inf, and -inf values
    df = df.dropna(axis=1, how='any')
    cond = ((df == float('inf')) | (df == float('-inf'))).any(axis=0)
    df = df.drop(df.columns[cond], axis=1)
    # Apply a simple variance threshold
    selector = VarianceThreshold()
    selector.fit(df)
    # Get only no-constant features
    top_features = selector.get_feature_names_out()
    df = df[top_features]
    return df

def max_num_in_list( list ):
    max = list[0]
    index = 0
    for i in range(1, len(list)):
        if list[i] > max:
            max = list[i]
            index = i
    return max, index

def min_num_in_list( list ):
    min = list[ 0 ]
    index = 0
    for i in range(1,len(list)):
        if list[i] < min:
            min = list[i]
            index = i
    return min,index


def getInitialIndex(dictTotal,cluster):
    listOfMinDist = list()
    listOfMinId = list()
    dictSumDistance = dict()
    for idX in dictTotal.keys():
        somma = 0
        for singleDict in dictTotal[idX]:
            somma += singleDict["distance"]
        dictSumDistance[idX]=somma
        if len(listOfMinDist) < cluster:
            listOfMinDist.append(dictSumDistance[idX])
            listOfMinId.append(idX)
        else:
            if max_num_in_list(listOfMinDist)[0] > somma:
                listOfMinDist[max_num_in_list(listOfMinDist)[1]] = dictSumDistance[idX]
                listOfMinId[listOfMinDist.index(dictSumDistance[idX])] = idX

    return listOfMinId


import pandas

import FeatTS.utilFeatExtr as util
from FeatTS.PFA import PFA
from concurrent.futures import ProcessPoolExecutor, as_completed
from multiprocessing import Manager
import pandas as pd
from tsfresh import feature_selection
import numpy as np
import time



class FeatTS(object):
    """
    	FeatTS method for time series clustering.

    	Parameters
    	----------
    	n_clusters : int, optional
    		Number of clusters (default is 2).

    	n_jobs : int, optional
    		Number of jobs tun run in parallel the graph computation for each length
    		(default is 4).

    	value_PFA : float, optional
    		Value of explained variance
    		(default is 0.9).

    	max_numb_feat : int, optional
    		Number of features max to adopt for the graph
    		(default is 20).

    	random_feat : bool, optional
    		Pick features in a random way and not based on importance
    		(default is False).

    	threshold_community : float, optional
    		Threshold of closeness between two values
    		(default is 0.8).

    	algorithm_community : ['Greedy','kClique','SLPA'], optional
    		Type of community detecion
    		(default is Greedy).
    	"""

    def __init__(self, n_clusters, n_jobs=1, value_PFA=0.9, max_numb_feat=20,
                 random_feat=False, threshold_community=0.8, algorithm_community='Greedy') :
        """
        initialize FeatTS method
        """
        self.n_clusters = n_clusters
        self.n_jobs = n_jobs
        self.value_PFA = value_PFA
        self.max_numb_feat = max_numb_feat
        self.random_feat = random_feat
        self.threshold_community = threshold_community
        self.algorithm_community = {algorithm_community:{}}
        self.feats_selected_ = []

    def fit(self, X, labels=None, external_feat: pd.DataFrame = None):
        """
        compute FeatTS on X

        Parameters
        ----------
        X : array of shape (n_samples, n_timestamps)
        	Training instances to cluster.

        y : dict of labels {idx:class}

        train_perc : percentage of semi-supervision (float)

        external_feat : features to use in combination with the features extracted (Dataframe)

        Returns
        -------
        self : object
        	Fitted estimator.
        	:param external_feat:
        	:param train_semi_supervised:
        """
        if external_feat is not None and X.shape[0] != external_feat.shape[0] :
            raise ValueError("The external features should have a feature value for each time series in input")

        if labels is not None:
            datasetAdapted = {"listOut": util.adaptTimeSeriesUCR(X),'labels': labels}

        else:
            datasetAdapted = {"listOut": util.adaptTimeSeriesUCR(X)}

        start_feat_sel = time.time()

        self.feats_selected_, features_filtered_direct = self.__features_extraction_selection(datasetAdapted, external_feat, self.value_PFA)

        print(f"Time to select features: {time.time() - start_feat_sel}")
        start_comm_detect = time.time()

        matrixNsym = self.__community_and_matrix_creation(self.feats_selected_, datasetAdapted, features_filtered_direct)

        print(f"Time to Detect Communities and Create Matrix: {time.time() - start_comm_detect}")
        start_clustering = time.time()
        self.labels_ = self.__cluster(matrixNsym, datasetAdapted)

        print(f"Time to cluster: {time.time() - start_clustering}")

    def __features_extraction_selection(self,datasetAdapted, external_feat, value_PFA):

        # Create the dataframe for the extraction of the features
        listOut = datasetAdapted["listOut"]

        features_filtered_direct = util.extractFeature(listOut, external_feat=external_feat, n_jobs=self.n_jobs)

        if external_feat is not None:
            external_feat = features_filtered_direct[external_feat.columns.tolist()].copy()
            # features_filtered_direct.drop(columns=external_feat.columns.tolist(), inplace=True)

        pfa = PFA()
        features_filtered_direct = util.cleaning(features_filtered_direct)

        if 'labels' in list(datasetAdapted.keys()):
            allAcc = list(datasetAdapted["labels"].keys())
            seriesAcc = pd.Series((datasetAdapted["labels"][i] for i in allAcc))
            filtreFeat = features_filtered_direct.loc[allAcc].reset_index(drop=True)

            multiclass = False
            significant_class = 1
            if len(seriesAcc.unique()) > 2:
                multiclass = True
                significant_class = len(seriesAcc.unique())


            if 'id' in filtreFeat.keys():
                filtreFeat = filtreFeat.drop('id', axis='columns')
            elif 'index' in filtreFeat.keys():
                filtreFeat = filtreFeat.drop('index', axis='columns')


            # Extract the relevance for each features and it will be ordered by importance
            ris = feature_selection.relevance.calculate_relevance_table(filtreFeat, seriesAcc,
                                                                        ml_task="classification",
                                                                        n_jobs=self.n_jobs,
                                                                        multiclass=multiclass,
                                                                        n_significant=significant_class)
            if external_feat is not None:
                ris = ris[~ris['feature'].isin(external_feat.columns.tolist())]

            if multiclass:
                p_value_columns = [col for col in ris.columns if col.startswith('p_value')]
                # Replace NaN values with inf in the p_value columns
                ris[p_value_columns] = ris[p_value_columns].fillna(np.inf)
                # Sum the p_value columns
                ris['p_value'] = ris[p_value_columns].sum(axis=1)

            ris = ris.sort_values(by='p_value')

            if self.random_feat:
                ris = util.randomFeat(ris, self.max_numb_feat)

            listOfFeatToUse = []
            for t in range(self.max_numb_feat):
                listOfFeatToUse.append(ris["feature"][t])

            dfFeatUs = pd.DataFrame()
            for x in range(len(listOfFeatToUse)):
                dfFeatUs[listOfFeatToUse[x]] = features_filtered_direct[listOfFeatToUse[x]]
            featPFA = pfa.fit(dfFeatUs, value_PFA)
        else:
            featPFA = pfa.fit(features_filtered_direct, value_PFA)

        if external_feat is not None:
            featPFA.extend(external_feat.columns.tolist())
            # Identify columns in external_feat that are not in features_filtered_direct
            non_overlapping_columns = external_feat.columns.difference(features_filtered_direct.columns)
            # Select only the non-overlapping columns from external_feat
            external_feat_non_overlapping = external_feat[non_overlapping_columns]
            # Perform the join with the non-overlapping columns
            features_filtered_direct = features_filtered_direct.join(external_feat_non_overlapping)

        return featPFA, features_filtered_direct

    def __community_and_matrix_creation(self, featPFA, datasetAdapted, features_filtered_direct):
        
        start_matrix_creation = time.time()
        listOfId = set(datasetAdapted["listOut"]["id"])
        manager = Manager()
        dictOfInfoTrain = manager.dict()
        dictOfInfoTrain = {}
        # Creation of the features that we want to use
        listOfFeat = featPFA

        def collect_result_Train(result):
            dictOfInfoTrain.update(result)


        chunk_size = len(listOfFeat) // self.n_jobs
        feature_chunks = [listOfFeat[i:i + chunk_size] for i in range(0, len(listOfFeat), chunk_size)]

        with ProcessPoolExecutor(max_workers=self.n_jobs) as executor:
            futures = [
                executor.submit(util.getCommunityDetectionTrain, feature_chunk, features_filtered_direct, listOfId,
                                self.threshold_community, self.n_clusters, self.algorithm_community)
                for feature_chunk in feature_chunks
            ]

            for future in as_completed(futures):
                collect_result_Train(future.result())

        print(f"Time to detect communities: {time.time() - start_matrix_creation}")

        
        setCluster = list()
        # Creation of list with all the cluster and their weights, used for the creation of CoOccurrence Matrix
        for key in dictOfInfoTrain.keys():
            for clusterInside in dictOfInfoTrain[key]["cluster"]:
                dictSing = {'list': list(clusterInside), 'weight': dictOfInfoTrain[key]["weightFeat"]}
                setCluster.append(dictSing)

        # Creation of CoOccurrence Matrix
        # print("Matrix Creation...")
        
        matrixNsym = util.getTabNonSym(setCluster, list(listOfId), n_jobs=self.n_jobs)
        print(f"Time to Create Matrix General: {time.time() - start_matrix_creation}")
        return matrixNsym

    def __cluster(self, matrixNsym, datasetAdapted):
        # print("Clustering Creation....")
        # List of the cluster created in the training set. It will be used later for the intersaction
        # with the cluster extract from the testing.
        listOfId = set(datasetAdapted["listOut"]["id"])

        listOfCommFindTest = util.getCluster(matrixNsym, listOfId, self.n_clusters)

        listOfCommFindTest = util.createSet(listOfCommFindTest, self.n_clusters)

        # Modify the index of the TimeSeries with their classes
        y_pred = [0 for x in range(len(listOfId))]
        for value in range(len(listOfCommFindTest)):
            for ind in listOfCommFindTest[value]["cluster"]:
                y_pred[ind] = value
        return y_pred


from collections import defaultdict

from sklearn.cluster import KMeans
from sklearn.metrics import euclidean_distances
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
import numpy as np

class PFA(object):
    def __init__(self, q=None):
        self.q = q

    def fit(self, X, expl_var = 0.9):
        if not self.q:
            self.q = X.shape[1]

        sc = StandardScaler()
        X_trans = sc.fit_transform(X)
        # Choice of the Explained Variance
        pca = PCA(expl_var)
        pca.fit(X_trans)
        princComp = len(pca.explained_variance_ratio_)
        A_q = pca.components_.T


        kmeans = KMeans(n_clusters=princComp).fit(A_q)
        clusters = kmeans.predict(A_q)
        cluster_centers = kmeans.cluster_centers_

        dists = defaultdict(list)
        for i, c in enumerate(clusters):
            dist = euclidean_distances([A_q[i, :]], [cluster_centers[c, :]])[0][0]
            dists[c].append((i, dist))

        self.indices_ = [sorted(f, key=lambda x: x[1])[0][0] for f in dists.values()]
        self.features_ = X_trans[:, self.indices_]
        listFeat = []
        for x in self.indices_:
            listFeat.append(X.columns[x])

        # print("Features Selected: " + str(listFeat))
        return listFeat



import numpy as np
import networkx as nx
from collections import defaultdict


def find_communities(G, T, r):
    """
    Speaker-Listener Label Propagation Algorithm (SLPA)
    see http://arxiv.org/abs/1109.5720
    """

    ##Stage 1: Initialization
    memory = {i: {i: 1} for i in G.nodes()}

    ##Stage 2: Evolution
    for t in range(T):
        listenersOrder = list(G.nodes())
        np.random.shuffle(listenersOrder)

        for listener in listenersOrder:
            speakers = G[listener].keys()
            if len(speakers) == 0:
                continue

            labels = defaultdict(int)

            for j, speaker in enumerate(speakers):
                # Speaker Rule
                total = float(sum(memory[speaker].values()))

                labels[list(memory[speaker].keys())[np.random.multinomial(1, [freq / total for freq in memory[speaker].values()]).argmax()]] += 1

            # Listener Rule
            acceptedLabel = max(labels, key=labels.get)

            # Update listener memory
            if acceptedLabel in memory[listener]:
                memory[listener][acceptedLabel] += 1
            else:
                memory[listener][acceptedLabel] = 1

    ## Stage 3:
    for node, mem in list(memory.items()):
        for label, freq in list(mem.items()):
            if freq / float(T + 1) < r:
                del mem[label]

    # Find nodes membership
    communities = {}
    for node, mem in list(memory.items()):
        for label in list(mem.keys()):
            if label in communities:
                communities[label].add(node)
            else:
                communities[label] = set([node])

    # Remove nested communities
    nestedCommunities = set()
    keys = list(communities.keys())
    for i, label0 in enumerate(keys[:-1]):
        comm0 = communities[label0]
        for label1 in keys[i + 1:]:
            comm1 = communities[label1]
            if comm0.issubset(comm1):
                nestedCommunities.add(label0)
            elif comm0.issuperset(comm1):
                nestedCommunities.add(label1)

    for comm in nestedCommunities:
        del communities[comm]

    return communities


from setuptools import setup, find_packages

__version__ = "0.0.4"

setup(
    name="FeatTS",
    version=__version__,
    description="A new method for clustering time series by adopting the best statistical features.",
    author="Donato Tiano",
    author_email="donatotiano@gmail.com",
    packages=find_packages(),
    zip_safe=True,
    license="",
    url="https://github.com/protti/FeatTS",
    entry_points={},
    install_requires=[
'fastdtw~=0.3.4',  
        'networkx~=2.5',  
        'numpy~=1.21',  
        'pandas~=1.3',  
        'pyclustering==0.10.1.2',  
        'scikit_learn~=0.24.2',  
        'scipy~=1.13.1',  
        'tsfresh~=0.18.0',  
        'aeon~=0.6.0'
        ]
)